/*
	$Id: AsciiConvertionMode.cpp,v 1.29 2003/05/14 12:03:14 phonohawk Exp $
	---------
	
    MacOS X implementation of the SKK input method.
    Copyright (C) 2002 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "AsciiConvertionMode.h"
#include "BIMInputEvents.h"
#include "HiraganaInputMode.h"
#include "ParentInputMode.h"
#include "CppCFData.h"
#include "WordRegisterMode.h"
#include "ServerConnectionFactory.h"
#include "KanjiConversionMode.h" // KanjiConversionMode̔񃁃o֐𗬗pB
#include "ClientConfiguration.h"
#include "SkkConfig.h"

AsciiConvertionMode::AsciiConvertionMode(ParentInputMode& parent)
	:ChildInputMode(parent),
	word_register_mode(NULL),
	status(STATUS_NULL),
	show_cands_window_after_Nth_cand(5),
	completion_mode(false),
    henkanModeStatus(false)
{

}

AsciiConvertionMode::~AsciiConvertionMode()
{
	if (word_register_mode) delete word_register_mode;
}

void AsciiConvertionMode::start()
{
	status = STATUS_WHITE;
}

bool AsciiConvertionMode::handleInput(unsigned char inputChar)
{
    bool result = handleInputChar(inputChar);
    henkanModeStatus = !result;
    
    return result;
}

bool AsciiConvertionMode::handleInputChar(unsigned char inputChar)
{	
	if (word_register_mode)
	{
		word_register_mode->handleInput(inputChar);
		return false; // WordRegisterMode::handleInput()ԂƂm͂ȂB
	}
	
	
	if (SkkConfig::config().isAbbrevNextKey(inputChar))
	{
		if (status == STATUS_WHITE && completion_mode)
		{
			// ̌ցBŌ܂ōsŏɖ߂B
			current_completion_index++;
			if (current_completion_index >= completions.size())
			{
				current_completion_index = 0;
			}
			index = completions[current_completion_index];
			return false;
		}
	}
	
	if (SkkConfig::config().isAbbrevPrevKey(inputChar))
	{
		if (status == STATUS_WHITE && completion_mode)
		{
			// ǑցBŏ܂ōsŌɖ߂B
			if (current_completion_index == 0)
			{
				current_completion_index = completions.size() - 1;
			}
			else
			{
				current_completion_index--;
			}
			index = completions[current_completion_index];
			return false;
		}
	}
	
	// tab,sIh,J}ȊÔȂ镶͂Ă⊮[h𔲂B
	completion_mode = false;
	
	if (SkkConfig::config().isToggleKanaKey(inputChar))
	{
		if (status == STATUS_BLACK || status == STATUS_BLACK_WITH_WINDOW)
		{
			// ݂̌mB
			// 1) [hȂSpЉ[h
			// 2) ȊOȂ畽[h
			
			parent->setEnabledKanjiMode(false);
			if (parent->isHiraganaInputMode()) {
				parent->goZenKataInputMode();
			}
			else {
				parent->goHiraganaInputMode();
			}
			return true;
		}
	}
	
	if (SkkConfig::config().isPrevKouhoKey(inputChar))
	{
		if (status == STATUS_BLACK)
		{
			// ǑցBŏ̌₾灤[h֖߂B
			if (current_candidate_index > 0)
			{
				current_candidate_index--;
			}
			else
			{
				status = STATUS_WHITE;
			}
			return false;
		}
		else if (status == STATUS_BLACK_WITH_WINDOW)
		{
			// deleteꂽ̂ƌ
			handleBackSpace();
			return false;
		}
		// ȊOȂfʂ
	}
	
	if (SkkConfig::config().isPurgeFromJisyoKey(inputChar))
	{
		if (status == STATUS_BLACK)
		{
			// 폜vvgoB̕ϊ̓[U[ɍڂĂȂƂĂ
			// ʓ|Ȃ̂ł̎킴킴mFȂBv]łΕʂB
			status = STATUS_PROMPT;
			prompt_type = PROMPT_DELETE_FROM_USER_DIC;
			prompt_input.clear();
			prompt.clear().append(0x672c).append(0x5f53).append(0x306b).append(' ');
			prompt.append('/').append(index).append('/');
			prompt.append(candidates[current_candidate_index]).append("/ ");
			prompt.append(0x3092).append(0x524a).append(0x9664).append(0x3057).append(0x3066).
				append(0x3082).append(0x5b9c).append(0x3057).append(0x3044).append(0x3067).
				append(0x3059).append(0x304b).append(0xff1f).append("(yes/no)");
			return false;
		}
	}
	
	if (SkkConfig::config().isToggleKatakanaKey(inputChar))
	{
		if (status == STATUS_WHITE)
		{
			// ppSppɕϊĂm肵ďI
			parent->fix(ZenAscInputMode::convert(index));
			
			parent->setEnabledAsciiConvertionMode(false);
			initialize();
			
			return false;
		}
		if (status == STATUS_BLACK || status == STATUS_BLACK_WITH_WINDOW)
		{
			// ݂̌m肵ĂA
			// 1) pЉ[hȂ畽[h
			// 2) ȊOȂ甼pЉ[h
			
			parent->setEnabledAsciiConvertionMode(false);
			if (parent->isHanKataInputMode()) {
				parent->goHiraganaInputMode();
			}
			else {
				parent->goHanKataInputMode();
			}
			return true;
		}
	}
	
	if (SkkConfig::config().isKakuteiKey(inputChar))
	{
		return true; // m̂
	}
	
	if (SkkConfig::config().isLatinModeKey(inputChar)
		|| SkkConfig::config().isJisx0208LatinModeKey(inputChar))
	{
		if (status == STATUS_BLACK)
		{
			// ꂼm肵Ĕp/Spp[hցB
			if (SkkConfig::config().isLatinModeKey(inputChar))
			{
				parent->goHanAscInputMode();
			}
			else
			{
				parent->goZenAscInputMode();
			}
			return true;
		}
		// [hIʂȂfʂB
	}
	
	if (SkkConfig::config().isAbbrevModeKey(inputChar))
	{
		if (status == STATUS_BLACK || status == STATUS_BLACK_WITH_WINDOW)
		{		
			parent->fix(this->getStringToFix());
			parent->handleInput(inputChar);
			return false;
		}
	}
	
	if (SkkConfig::config().isNextKouhoKey(inputChar))
	{
		if (status == STATUS_WHITE)
		{
			if (index.length() == 0)
			{
				// I
				parent->setEnabledAsciiConvertionMode(false);
				initialize();
				return false;
			}
			
			// Iɖ₢킹
			askServerTheCandidates();
			
			if (candidates.size() > 0)
			{
				// ₪݂B[hցB
				status = STATUS_BLACK;
				current_candidate_index = 0; // ŏ̌B
			}
			else
			{
				// ₪BPo^JnB
				startRegisteringWord();
			}
		}
		else if (status == STATUS_BLACK)
		{
			// ̌
			current_candidate_index++;
			
			if (current_candidate_index >= candidates.size()) // I܂ōsPo^JnB
			{
				startRegisteringWord();
			}
			else if (current_candidate_index >= show_cands_window_after_Nth_cand - 1) // ȍ~̃[e[VŌIEChE\B
			{
				status = STATUS_BLACK_WITH_WINDOW;
				openCandidatesWindow();
			}
		}
		else if (status == STATUS_BLACK_WITH_WINDOW)
		{
			if (candidates_window_current_frame < candidates_window_num_of_frames - 1)
			{
				// Ō̃t[łȂ΁Ãt[֍sB
				goToNextCandidatesFrame();
			}
			else
			{
				// Ō̃t[APo^JnB
				closeCandidatesWindow();
				startRegisteringWord();
			}
		}
	}
	else
	{
		if (status == STATUS_WHITE)
		{
			// indexɂǉB
			index += inputChar;
		}
		else if (status == STATUS_BLACK)
		{
			// ݂̌m肵AďI邪A
			// ɍ̕\/fixfȂƍ̂
			// AXL[ϊ[hIĂparenthandleInputɓnB
			// AsciiConvertionModeSTATUS_NULL̎ɂ
			// kIgnoreThisԂB
			
			// IɑM
			sendCurrentCandidateToServerToRegister();
			
			// EȂAIɊm肳B
			parent->fix(candidates[current_candidate_index]);
			
			parent->setEnabledAsciiConvertionMode(false);
			initialize();
			
			parent->handleInput(inputChar);
		}
		else if (status == STATUS_BLACK_WITH_WINDOW)
		{
			// asdfjkli1t[̌␔7菭Ȃ΂Bj̕ł
			// ̌m肷B
			// łȂΌ݂̌m肷B
			int cand_index = CppCFString("asdfjkl").substring(0,candidates_window_cands_per_frame).indexOf(inputChar);
			if (cand_index != -1)
			{
				unsigned new_index = current_candidate_index + cand_index;
				if (new_index < candidates.size()) {
					// ̃CfBbNXȂmB
					current_candidate_index = new_index;
					return true;
				}				
			}
			return false;
		}
		else if (status == STATUS_PROMPT)
		{
			// ̂܂prompt_inputɒǉB
			prompt_input += inputChar;
		}
	}
	return false;
}

void AsciiConvertionMode::startRegisteringWord()
{
	word_register_mode = new WordRegisterMode(*this,index);
	word_register_mode->redisplay();
}

bool AsciiConvertionMode::handleBackSpace()
{
	if (word_register_mode)
	{
		word_register_mode->handleBackSpace();
		return true; // WordRegisterMode::handleBackSpace()ԂƂ{͏ȂB
	}
	
	if (status == STATUS_NULL)
	{
		return false;
	}
	
	if (status == STATUS_WHITE)
	{
		// indexȂ珉ďI
		// łȂindex̍Ō̕B
		if (index.length() == 0)
		{
			parent->setEnabledAsciiConvertionMode(false);
			parent->display(CppCFString());
			initialize();
		}
		else
		{
			completion_mode = false; // ⊮[hꍇ͕⊮[h𔲂B
			index.eraseLast(1);
		}
	}
	else if (status == STATUS_BLACK)
	{
		// ݂̌m肵ĂŌ̕폜B
		// ďďIB
		CppCFString str_to_fix(candidates[current_candidate_index]);
		str_to_fix.eraseLast(1);
		
		// IɑM
		sendCurrentCandidateToServerToRegister();
		
		// ̎
		parent->fix(str_to_fix);
		parent->display(CppCFString());
		
		parent->setEnabledAsciiConvertionMode(false);
		initialize();
	}
	else if (status == STATUS_BLACK_WITH_WINDOW)
	{
		// ŏ̃t[ȂA[h֖߂
		// łȂΑÕt[ցB
		if (candidates_window_current_frame == 0)
		{
			// EChEJOɖ߂
			current_candidate_index = show_cands_window_after_Nth_cand - 2;
			
			// EChEB
			closeCandidatesWindow();
			
			status = STATUS_BLACK;
		}
		else
		{
			goToPrevCandidatesFrame();
		}
	}
	else if (status == STATUS_PROMPT)
	{
		// prompt_inputłȂ΍Ō̕B
		if (prompt_input.length() > 0)
		{
			prompt_input.eraseLast(1);
		}
	}
	return true;
}

bool AsciiConvertionMode::handleReturn()
{
	if (word_register_mode)
	{
		word_register_mode->handleReturn();
		
		if (word_register_mode) // handleReturnwordRegistrationFinishedĂ΂ꂽword_register_mode͔j邽߁B
		{
			word_register_mode->redisplay();
		}
		
		return true; // trueԂB
	}
	
	if (status == STATUS_PROMPT)
	{
		switch (prompt_type)
		{
			case PROMPT_DELETE_FROM_USER_DIC:
			{
				// prompt_inputyesȂ폜STATUS_WHITEɖ߂B
				// ȊOȂSTATUS_BLACKɖ߂B
				if (prompt_input == "yes")
				{
					// 폜
					makeServerRemoveWord();
					
					// STATUS_WHITE
					status = STATUS_WHITE;
				}
				else
				{	
					status = STATUS_BLACK;
				}
			}
			break;
			
			default:
			break;
		}
		return true;
	}
	
	if (ClientConfiguration::sharedConfiguration().isSkkEggLikeNewline())
	{
		// skk-egglike-newline[hȂA͂Ŋm肳Ă珉trueԂB
		
		// IɑM
		parent->fix(getStringToFix());
		return true;
	}
	else
	{
		return false;
	}
}

bool AsciiConvertionMode::handleArrow(int kind)
{
	if (word_register_mode) {
		return word_register_mode->handleArrow(kind); // Ϗ
	}
	
	if (status == STATUS_BLACK || status == STATUS_BLACK_WITH_WINDOW)
	{
		// IɑM
		sendCurrentCandidateToServerToRegister();
		
		// ݂̌mB
		parent->fix(candidates[current_candidate_index]);
		
		if (status == STATUS_BLACK_WITH_WINDOW)
		{
			// EChEB
			closeCandidatesWindow();
		}
	}
	
	// AXL[ϊ[hIB
	parent->setEnabledAsciiConvertionMode(false);
	initialize();
	return false;
}

bool AsciiConvertionMode::handleCg()
{
	if (word_register_mode)
	{
		return word_register_mode->handleCg();
	}
	
	// Cvbg[hhandleCgĂсAgetStringToFix()ĂԂAĂԂB
	parent->getCurrentInputMode().handleCg();
	parent->getCurrentInputMode().getStringToFix();
	
	if (status == STATUS_WHITE)
	{
		// AXL[ϊ[hI
		parent->setEnabledAsciiConvertionMode(false);
		initialize();
	}
	else if (status == STATUS_BLACK || status == STATUS_BLACK_WITH_WINDOW)
	{
		// okuriȂSTATUS_WHITE_ROOTցBłȂSTATUS_WHITE_OKURIցB
		// EChEJĂB
		if (status == STATUS_BLACK_WITH_WINDOW)
		{
			closeCandidatesWindow();
		}
		
		status = STATUS_WHITE;
	}
	return true;
}

bool AsciiConvertionMode::handleCy()
{
	if (word_register_mode)
	{
		return word_register_mode->handleCy();
	}
	else {
		return false;
	}
}

bool AsciiConvertionMode::handleTab()
{
	if (word_register_mode)
	{
		return word_register_mode->handleTab();
	}
	
	if (status == STATUS_WHITE)
	{
		if (completion_mode)
		{
			// ̌ցBŌ܂ōsŏɖ߂B
			current_completion_index++;
			if (current_completion_index >= completions.size())
			{
				current_completion_index = 0;
			}
			index = completions[current_completion_index];
		}
		else
		{
			// Iɖ₢킹B
			askServerTheCompletions();
			
			// ⊮̌₪Ε⊮JnBΉɖ߂B
			if (completions.size() > 0)
			{
				completion_mode = true;
				index = completions[0];
			}
		}
		return false;
	}
	
	return false;
}

void AsciiConvertionMode::terminate()
{
	if (word_register_mode)
	{
		word_register_mode->terminate();
	}
	
	parent->fix(this->getStringToFix()); // m肷
	parent->setEnabledAsciiConvertionMode(false);
	initialize();
}

void AsciiConvertionMode::wordRegistrationFinished()
{
	if (word_register_mode)
	{
		CppCFString content = word_register_mode->getContent();
		if (content.length() > 0)
		{
			// IɑM
			sendWordToServerToRegister(index,content);
			
			// m
			parent->fix(content);
			
			// I
			parent->setEnabledAsciiConvertionMode(false);
			initialize();
		}
		else
		{
			status = STATUS_WHITE;
		}
		
		delete word_register_mode;
		word_register_mode = NULL;
		parent->updatePencilMenuIcon();
	}
}

void AsciiConvertionMode::wordRegistrationCanceled()
{
	delete word_register_mode;
	word_register_mode = NULL;
	parent->updatePencilMenuIcon();
	
	if (status == STATUS_WHITE)
	{
		// do nothing
	}
	else if (status == STATUS_BLACK)
	{
		current_candidate_index = candidates.size() - 1; // Ō̌ցB
	}
	else if (status == STATUS_BLACK_WITH_WINDOW)
	{
		// EChEJčŌ̃t[o
		current_candidate_index = show_cands_window_after_Nth_cand - 1; // goToNextCandidatesFrame()łB
		openCandidatesWindow();
		for (unsigned i = 1;i < candidates_window_num_of_frames;i++)
		{
			goToNextCandidatesFrame();
		}
	}
}

CppCFString AsciiConvertionMode::getStringToDisplay()
{
	if (word_register_mode)
	{
		// BȂƐ܊pWordRegisterModeݒ肵ׂĂ܂B
		return CppCFString().append(kIgnoreThis);
	}
	
	if (status == STATUS_WHITE)
	{
		// index
		return CppCFString().append(0x25bd).append(index);
	}
	else if (status == STATUS_BLACK || status == STATUS_BLACK_WITH_WINDOW)
	{
		// ݂̌
		return CppCFString().append(0x25bc).append(candidates[current_candidate_index]);
	}
	else if (status == STATUS_PROMPT)
	{
		// [vvg][]
		return CppCFString('[').append(prompt).append("][").append(prompt_input).append(kCaret).append(']');
	}
	else
	{
		return CppCFString().append(kIgnoreThis);
	}
}

CppCFString AsciiConvertionMode::getStringToFix()
{
	CppCFString result;
	if (status == STATUS_WHITE)
	{
		result = index;
	}
	else if (status == STATUS_BLACK)
	{
		result = candidates[current_candidate_index];
		
		// IɑM
		sendCurrentCandidateToServerToRegister();
	}
	else if (status == STATUS_BLACK_WITH_WINDOW)
	{
		result = candidates[current_candidate_index];
		
		// IɑM
		sendCurrentCandidateToServerToRegister();
		
		// EChEB
		closeCandidatesWindow();
	}
	else// if (status == STATUS_NULL)
	{
		result = CppCFString().append(kIgnoreThis);
	}
	
	// ꂪĂ΂ꂽƂƂfixƂƂȂ̂ŃAXL[ϊ[hIB
	// B
	parent->setEnabledAsciiConvertionMode(false);
	initialize();
	
	return result;
}

void AsciiConvertionMode::initialize()
{
	status = STATUS_NULL;
	index.clear();
	candidates.clear();
	current_candidate_index = 0;
	completion_mode = false;
	completions.clear();
	current_completion_index = 0;
}

void AsciiConvertionMode::askServerTheCandidates()
{
	// ahya@@-ahya
	CppCFData query(CppCFString().append('-').append(index).toCFData());
	
	current_candidate_index = 0;
	::askServerTheCandidates(query,candidates);
}

void AsciiConvertionMode::askServerTheCompletions()
{	
	// Iɐq˂
	CppCFData cfdata_query(index.toCFData());
	current_completion_index = 0;
	::askServerTheCompletions(cfdata_query,completions);
}

void AsciiConvertionMode::sendCurrentCandidateToServerToRegister()
{
	// IɑM
	sendWordToServerToRegister(index,candidates[current_candidate_index]);
}

void AsciiConvertionMode::makeServerRemoveWord()
{
	CppCFString query;
	// ahya i߁߁j@@-ahya i߁߁j
	query.append('-').append(index).append(' ').append(candidates[current_candidate_index]);
	
	// Iɓn
	CppCFData cfdata_query(query.toCFData());
	::makeServerRemoveWord(cfdata_query);
}

void AsciiConvertionMode::openCandidatesWindow()
{
	// candidates4Ԗڈȍ~̍ڂzB
	//vector<CppCFString> cands(candidates.begin()+(show_cands_window_after_Nth_cand-1),candidates.end());
	CppCFString space(" ");
	CppCFString nbsp("[20]");
	vector<CppCFString> cands;
	for (vector<CppCFString>::const_iterator ite = candidates.begin() + (show_cands_window_after_Nth_cand-1);ite != candidates.end();ite++)
	{
		cands.push_back(ite->replaceClone(space,nbsp));
	}
	
	::openCandidatesWindow(join(' ',cands),
							candidates_window_id,
							candidates_window_cands_per_frame,
							candidates_window_num_of_frames);
	candidates_window_current_frame = 0;
}


void AsciiConvertionMode::goToNextCandidatesFrame()
{
	// EChEID(UInt16)܂CFData쐬AI֑B
	ServerConnectionFactory::sharedFactory().newConnection().send(kSKKCandidatesWindowNext,newCFDataRefWithWindowID(candidates_window_id));
	
	// ݂̃t[ԍCNg
	candidates_window_current_frame++;
	
	// ݂̌ԍt[̌␔₷B
	current_candidate_index += candidates_window_cands_per_frame;
}

void AsciiConvertionMode::goToPrevCandidatesFrame()
{
	// EChEID(UInt16)܂CFData쐬AI֑B
	ServerConnectionFactory::sharedFactory().newConnection().send(kSKKCandidatesWindowPrev,newCFDataRefWithWindowID(candidates_window_id));
	
	// ݂̃t[ԍfNg
	candidates_window_current_frame--;
	
	// ݂̌ԍt[̌␔炷B
	current_candidate_index -= candidates_window_cands_per_frame;
}

void AsciiConvertionMode::closeCandidatesWindow()
{
	// EChEID(UInt16)܂CFData쐬AI֑B
	ServerConnectionFactory::sharedFactory().newConnection().send(kSKKCloseCandidatesWindow,newCFDataRefWithWindowID(candidates_window_id));
}

