/*
	$Id: CharacterTranslator.cpp,v 1.4 2003/04/21 14:23:48 phonohawk Exp $
	---------
	
    MacOS X implementation of the SKK input method.
    Copyright (C) 2002 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "CharacterTranslator.h"
#include <fstream>
#include <Carbon/Carbon.h>
#include "BIMClientServer.h"
#include "SkkConfig.h"

CharacterTranslator::CharacterTranslator(const string& table_name)
{
	// ܂~/Library/AquaSKK/${table_name}ǂ߂邩ǂ𒲂ׂB
	string user_table_file = SkkConfig::home() + "/Library/AquaSKK/" + table_name;
	ifstream ifs(user_table_file.c_str(),ios::in);
	if (!ifs.fail())
	{
		// ǂݍ݉\B
		load(ifs);
	}
	else
	{
		CFBundleRef component_bundle = CFBundleGetBundleWithIdentifier(CFSTR(kIMBundleIdentifier));
		
		// ftHge[u`t@Cւ̃pXURLƂč쐬
		CFURLRef tablefile_url = NULL;
		if (component_bundle) {
			tablefile_url = CFBundleCopyResourceURL(component_bundle,CppCFString(table_name.c_str()).getString(),NULL,NULL);
		}
		
		if (tablefile_url) {
			// URLtpXAt@CJă[hB
			CFStringRef path_cfstr = CFURLCopyFileSystemPath(tablefile_url,kCFURLPOSIXPathStyle);
			
			CppCFString path(path_cfstr);
			ifstream is(path.toCString(),ios::in);
			if (!is.fail())
			{
				load(is);
			}
			
			CFRelease(path_cfstr);
		}
		
		if (tablefile_url) CFRelease(tablefile_url);
	}
}

CharacterTranslator::~CharacterTranslator()
{

}

void CharacterTranslator::load(istream& is)
{	
	string str_line;
	while (!is.eof())
	{
		getline(is,str_line,'\n');
		if (str_line.length() == 0) continue;
		if (str_line[0] == '#') continue;
		
		CppCFString line(str_line.c_str(),kCFStringEncodingEUC_JP);
		vector<CppCFString> items = line.split(',');
		
		if (items.size() != 2)
		{
			// ُȍsB
			printf("AquaSKK-CharacterTranslator: Illegal declaration of table.\n");
			printf("   %s\n",str_line.c_str());
			continue;
		}
		
		CppCFString from = items[0].replace("&comma;",",");
		CppCFString to = items[1].replace("&comma;",",");
		
		if (from.length() == 0)
		{
			// ُȍsB
			printf("AquaSKK-CharacterTranslator: Illegal declaration of table.\n");
			printf("   %s\n",str_line.c_str());
			continue;
		}
		
		// from̕āAׂvector̗vf݂邩ǂ𒲂ׁA΍B
		if (from.length() > table_root.size())
		{
			table_root.resize(from.length());
			// uresize()ŃTCYĂ̈͊JȂBvƂ邪A
			// ꂪ{ƂĂǂCB
		}
		
		pair<CppCFString,CppCFString> current_pair(from,to);
		table_root[from.length() - 1].push_back(current_pair);
	}
}

CppCFString CharacterTranslator::translate(const CppCFString& src)
{
	CppCFString dest;
	
	const unsigned table_root_size = table_root.size();
	const unsigned srclen = src.length();
	for (unsigned pos = 0;pos < srclen;)
	{
		// ̑̂珇ɒׂĂBȂƐuȂB
		bool replaced = false;
		for (unsigned find_len = table_root_size;find_len >= 1;find_len--)
		{
			vector<pair<CppCFString,CppCFString> >& pairs = table_root[find_len - 1];
			for (vector<pair<CppCFString,CppCFString> >::const_iterator ite = pairs.begin();ite != pairs.end();ite++)
			{
				if (src.startsWith(ite->first,pos)) // t
				{
					// destsecondRs[
					dest += ite->second;
					
					// posfirst̕i߂B
					pos += find_len;
					
					// replacedtO𗧂ĂāA[v𔲂B
					replaced = true;
					break;
				}
			}
			
			if (replaced) break; // ̃[vŒuꂽ炱̃[vB
		}
		
		if (!replaced) // uĂȂ
		{
			// pos̈ʒüꕶ̂܂܃Rs[
			dest += src[pos];
			pos++;
		}
	}
	
	return dest;
}
