/*
    $Id: CppCFString.cpp,v 1.15 2003/06/22 10:56:40 phonohawk Exp $
    ---------
	
    MacOS X implementation of the SKK input method.
    Copyright (C) 2002 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "CppCFString.h"
#include <string.h>

CppCFString::CppCFString()
{
	cf_string = CFStringCreateMutable(NULL,0);
}

CppCFString::CppCFString(const CppCFString& cpp_cf_string_src)
{
	// Rs[RXgN^
	cf_string = CFStringCreateMutableCopy(NULL,0,cpp_cf_string_src.getString());
}

CppCFString::CppCFString(UniChar c)
{
	cf_string = CFStringCreateMutable(NULL,0);
	append(c);
}

CppCFString::CppCFString(const char* c_string)
{
	cf_string = CFStringCreateMutable(NULL,0);
	
	unsigned len = strlen(c_string);
	for (unsigned i = 0;i < len;i++)
	{
		append(c_string[i]);
	}
}

CppCFString::CppCFString(const char* c_string,CFStringEncoding encoding)
{
	cf_string = CFStringCreateMutable(NULL,0);
	CFStringAppendCString(cf_string,c_string,encoding);
}

CppCFString::CppCFString(CFStringRef cf_string_src)
{
	cf_string = CFStringCreateMutableCopy(NULL,0,cf_string_src);
}

CppCFString::CppCFString(const UniChar* uni_string,unsigned int uni_string_len)
{
	cf_string = CFStringCreateMutable(NULL,0);
	CFStringAppendCharacters(cf_string,uni_string,uni_string_len);
}

CppCFString::CppCFString(CFDataRef cfdata)
{
	// CFDataRefUniChar̔złȂ΂ȂȂB
	const int size = CFDataGetLength(cfdata);
	const int length = size / sizeof(UniChar);
	UniChar* buf = new UniChar[length];
	
	CFDataGetBytes(cfdata,CFRangeMake(0,size),reinterpret_cast<UInt8*>(buf));
	cf_string = CFStringCreateMutable(NULL,0);
	CFStringAppendCharacters(cf_string,buf,length);
	delete buf;
}
		
CppCFString::~CppCFString()
{
	CFRelease(cf_string);
}

CppCFString& CppCFString::append(const CppCFString& str)
{
	CFStringAppend(cf_string,str.getString());
	return *this;
}
 
CppCFString& CppCFString::append(const UniChar c)
{
	CFStringAppendCharacters(cf_string,&c,1);
	return *this;
}

CppCFString& CppCFString::insert(unsigned offset,UniChar c)
{
	CFStringInsert(cf_string,offset,CppCFString(c).getString());
	return *this;
}

CppCFString& CppCFString::insert(unsigned offset,const CppCFString& str)
{
	CFStringInsert(cf_string,offset,str.getString());
	return *this;
}

CppCFString& CppCFString::clear()
{
	CFStringDelete(cf_string,CFRangeMake(0,CFStringGetLength(cf_string)));
	return *this;
}

unsigned int CppCFString::length() const
{
	return CFStringGetLength(cf_string);
}

CppCFString& CppCFString::erase(unsigned start_pos,unsigned end_pos)
{
	// ̓`FbNȂBӁB
	CFStringDelete(cf_string,CFRangeMake(start_pos,end_pos - start_pos));
	return *this;
}

CppCFString& CppCFString::eraseCharAt(unsigned pos)
{
	// ̓`FbNȂBӁB
	CFStringDelete(cf_string,CFRangeMake(pos,1));
	return *this;
}

CppCFString& CppCFString::eraseFirst(const unsigned int n_chars)
{
	// ̓`FbNȂBӁB
	CFStringDelete(cf_string,CFRangeMake(0,n_chars));
	return *this;
}

CppCFString& CppCFString::eraseLast(const unsigned int n_chars)
{
	// ̓`FbNȂBӁB
	CFStringDelete(cf_string,CFRangeMake(CFStringGetLength(cf_string) - n_chars,n_chars));
	return *this;
}

CppCFString CppCFString::substring(unsigned int start_pos) const
{
	return substring(start_pos,length());
}

CppCFString CppCFString::substring(const unsigned int start_pos,const unsigned int end_pos) const
{
	// start_pos͊܂ނend_pos͊܂܂ȂB
	// E`FbN͂Ȃ̂ŒӁB
	const UniChar* unistr = CFStringGetCharactersPtr(cf_string);
	if (unistr != NULL)
	{
		return CppCFString(&unistr[start_pos],end_pos - start_pos);
	}
	else
	{
		unsigned len = end_pos - start_pos;
		
		if (len > 0)
		{
			UniChar* buffer = new UniChar[len];
			
			CFStringGetCharacters(cf_string,CFRangeMake(start_pos,len),buffer);
			CppCFString result(buffer,len);
			
			delete buffer;
			return result;
		}
		else
		{
			return CppCFString();
		}
	}
}

CppCFString CppCFString::trim() const {
    // n܂ƏI͂0x20ȉ̕SďAVɐĕԂB
    if (length() == 0) {
	return CppCFString();
    }

    unsigned int len = length();
    unsigned int begin;
    for (begin = 0; begin < len; begin++) {
	if (CFStringGetCharacterAtIndex(cf_string, begin) > 0x20) {
	    break;
	}
	else if (begin == len-1) {
	    return CppCFString();
	}
    }
    int end;
    for (end = len-1; end >= 0; end--) {
	if (CFStringGetCharacterAtIndex(cf_string, end) > 0x20) {
	    break;
	}
	else if (end == 0) {
	    return CppCFString();
	}
    }

    return substring(begin, end + 1);
}

int CppCFString::indexOf(UniChar c,unsigned int search_start_pos) const
{
	const int len = CFStringGetLength(cf_string);
	UniChar* unistr = const_cast<UniChar*>(CFStringGetCharactersPtr(cf_string));
	bool unistr_is_independent = false;
	if (unistr == NULL)
	{
		unistr = new UniChar[len];
		unistr_is_independent = true;
		
		CFStringGetCharacters(cf_string,CFRangeMake(0,len),unistr);
	}
	
	int result = -1;
	for (int i = search_start_pos;i < len;i++)
	{
		if (unistr[i] == c)
		{
			result = i;
			break;
		}
	}
	
	if (unistr_is_independent)
	{
		delete unistr;
	}
	return result;
}

bool CppCFString::startsWith(const CppCFString& str,int search_start_pos) const
{
	if (length() - static_cast<unsigned>(search_start_pos) < str.length()) return false; // ĂȂB
	return substring(search_start_pos,search_start_pos + str.length()) == str;
}

bool CppCFString::endsWith(const CppCFString& str,int search_start_pos) const
{
	if (search_start_pos == -1) search_start_pos = length()-1;
	
	if (static_cast<unsigned>(search_start_pos) < str.length() - 1) return false; // ĂȂB
	return substring(search_start_pos+1 - str.length(),search_start_pos+1) == str;
}

CppCFString& CppCFString::replace(const CppCFString& from,const CppCFString& to)
{
	while (true)
	{
		CFRange found_range = CFStringFind(cf_string,from.getString(),0);
		if (found_range.location == kCFNotFound) break; // I
		
		CFStringReplace(cf_string,found_range,to.getString());
	}
	return *this;
}

CppCFString CppCFString::replaceClone(const CppCFString& from,const CppCFString& to) const
{
	return clone().replace(from,to);
}

UniChar CppCFString::operator[] (const unsigned int index) const
{
	return CFStringGetCharacterAtIndex(cf_string,index);
}

/*UniChar& CppCFString::operator[] (const unsigned int index)
{
	// {ɂȂƂėǂ̂낤c
	// ʖڂȂSTL̃Ce[^̂悤Ȃ̂邵ȂB
	UniChar* unistring_ptr = const_cast<UniChar*>(CFStringGetCharactersPtr(cf_string));
	return unistring_ptr[index];
}*/

CppCFString& CppCFString::operator= (const CppCFString& str)
{
	CFStringDelete(cf_string,CFRangeMake(0,CFStringGetLength(cf_string)));
	CFStringAppend(cf_string,str.getString());
	return *this;
}

CppCFString& CppCFString::operator+= (const CppCFString& str)
{
	append(str);
	return *this;
}

CppCFString& CppCFString::operator+= (const UniChar c)
{
	append(c);
	return *this;
}

CFStringRef CppCFString::getString() const
{
	return cf_string;
}

UniChar* CppCFString::toUniCharPtr() const
{
	// Ԃꂽ|C^́AKvȂdelete[]Ȃ΂Ȃ܂B
	const int len = CFStringGetLength(cf_string);
	UniChar* unistr = new UniChar[len];
	CFStringGetCharacters(cf_string,CFRangeMake(0,len),unistr);
	
	return unistr;
}

CFDataRef CppCFString::toCFData() const
{
	// ԂꂽCFDataRef́AgpɕKCFRelease邱ƁBȂ΃[NB
	const int len = CFStringGetLength(cf_string);
	UniChar* unistr = const_cast<UniChar*>(CFStringGetCharactersPtr(cf_string));
	bool unistr_is_independent = false;
	if (unistr == NULL)
	{
		unistr = new UniChar[len];
		unistr_is_independent = true;
		
		CFStringGetCharacters(cf_string,CFRangeMake(0,len),unistr);
	}
	
	unsigned int size = len * sizeof(UniChar);
	
	CFMutableDataRef result = CFDataCreateMutable(NULL,size);
	CFDataAppendBytes(result,
			reinterpret_cast<const UInt8*>(unistr),
			size);
	
	if (unistr_is_independent)
	{
		delete unistr;
	}
	return result;
}

char* CppCFString::toCString(CFStringEncoding encoding) const
{
	// Ԃꂽchar*́AgpɕKdelete[]邱ƁBȂ΃[NB
	const int len = CFStringGetLength(cf_string);
	char* buf = NULL;
	
	if (len == 0)
	{
		buf = new char[1];
		buf[0] = 0;
	}
	else
	{
		int ratio = 3;
		do
		{
			if (buf) delete buf;
			buf = new char[len * ratio];
			ratio++;
		} while (!CFStringGetCString(cf_string,buf,len * ratio,encoding));
	}
	return buf;
}

CppCFString CppCFString::clone() const
{
	return CppCFString(*this);
}

vector<CppCFString> CppCFString::split(UniChar delim) const
{
	const int len = CFStringGetLength(cf_string);
	UniChar* unistr = const_cast<UniChar*>(CFStringGetCharactersPtr(cf_string));
	bool unistr_is_independent = false;
	if (unistr == NULL)
	{
		unistr = new UniChar[len];
		unistr_is_independent = true;
		
		CFStringGetCharacters(cf_string,CFRangeMake(0,len),unistr);
	}
	
	vector<CppCFString> result;
	
	int pos = 0;
	while (pos < len)
	{
		if (unistr[pos] == delim) // f~^B
		{
			pos++;
		}
		else // f~^łȂtB
		{
			int blockstart = pos;
			int blockend = -1;
			
			for (int i = blockstart;i < len;i++)
			{
				// f~^TB
				if (unistr[i] == delim)
				{
					blockend = i;
					break;
				}
			}
			
			if (blockend == -1) // f~^tȂ
			{
				blockend = len; // Ō܂łg[NƂB
			}
			
			result.push_back(substring(blockstart,blockend));
			pos = blockend;
		}
	}
	
	if (unistr_is_independent)
	{
		delete unistr;
	}
	return result;
}

// 񃁃o֐Q
CppCFString operator+ (const CppCFString& l,const CppCFString& r)
{
	return CppCFString(l).append(r);
}

CppCFString operator+ (const CppCFString& l,const UniChar r)
{
	return CppCFString(l).append(r);
}

CppCFString operator+ (const UniChar l,const CppCFString& r)
{
	return CppCFString().append(l).append(r);
}

bool operator== (const CppCFString& l,const CppCFString& r)
{
	/*CFStringRef l_str = l.getString();
	CFStringRef r_str = r.getString();
	int l_len = CFStringGetLength(l_str);
	int r_len = CFStringGetLength(r_str);
	
	if (l_len != r_len) return false; // ȂΈvĂȂ̂炩B
	
	UniChar* l_unistr = const_cast<UniChar*>(CFStringGetCharactersPtr(l_str));
	bool l_unistr_is_independent = false;
	if (l_unistr == NULL)
	{
		l_unistr = new UniChar[l_len];
		l_unistr_is_independent = true;
		
		CFStringGetCharacters(l_str,CFRangeMake(0,l_len),l_unistr);
	}
	
	UniChar* r_unistr = const_cast<UniChar*>(CFStringGetCharactersPtr(r_str));
	bool r_unistr_is_independent = false;
	if (r_unistr == NULL)
	{
		r_unistr = new UniChar[r_len];
		r_unistr_is_independent = true;
		
		CFStringGetCharacters(r_str,CFRangeMake(0,r_len),r_unistr);
	}
	
	bool mismatched = false;
	UniChar* l_unistr_end = l_unistr + l_len * sizeof(UniChar);
	for (;l_unistr != l_unistr_end;l_unistr += sizeof(UniChar) , r_unistr += sizeof(UniChar))
	{
		if (*l_unistr != *r_unistr)
		{
			mismatched = true;
			break;
		}
	}
	
	if (l_unistr_is_independent) delete l_unistr;
	if (r_unistr_is_independent) delete r_unistr;
	return !mismatched;*/
	
	return CFStringCompare(l.getString(),r.getString(),0) == kCFCompareEqualTo;
}

bool operator!= (const CppCFString& l,const CppCFString& r)
{
	//return !(l == r);
	
	return CFStringCompare(l.getString(),r.getString(),0) != kCFCompareEqualTo;
}

bool operator< (const CppCFString& l,const CppCFString& r)
{
	return CFStringCompare(l.getString(),r.getString(),0) == kCFCompareLessThan;
}

bool operator> (const CppCFString& l,const CppCFString& r)
{
	return CFStringCompare(l.getString(),r.getString(),0) == kCFCompareGreaterThan;
}

bool operator<= (const CppCFString& l,const CppCFString& r)
{
	CFComparisonResult comp = CFStringCompare(l.getString(),r.getString(),0);
	return (comp == kCFCompareEqualTo || comp == kCFCompareLessThan);
}

bool operator>= (const CppCFString& l,const CppCFString& r)
{
	CFComparisonResult comp = CFStringCompare(l.getString(),r.getString(),0);
	return (comp == kCFCompareEqualTo || comp == kCFCompareGreaterThan);
}

CppCFString join(UniChar delim,const vector<CppCFString>& vec)
{
	CppCFString result;
	
	for (vector<CppCFString>::const_iterator ite = vec.begin();ite != vec.end();ite++)
	{
		result += *ite + delim;
	}
	
	if (result.length() > 0)
		return result.eraseLast(1);
	else
		return result;
}

// hash_CppCFString̃\bh
size_t hash_CppCFString::operator() (const CppCFString& str) const
{
	CFStringRef cf_string = str.getString();
	int len = CFStringGetLength(cf_string);
	UniChar* unistr = const_cast<UniChar*>(CFStringGetCharactersPtr(cf_string));
	bool unistr_is_independent = false;
	if (unistr == NULL)
	{
		unistr = new UniChar[len];
		unistr_is_independent = true;
		
		CFStringGetCharacters(cf_string,CFRangeMake(0,len),unistr);
	}
	
	unsigned long result = 0;
	UniChar* unistr_end = unistr + len * sizeof(UniChar);
	for (;unistr != unistr_end;unistr += sizeof(UniChar))
	{
		result = 5 * result + *unistr;
	}
	
	if (unistr_is_independent) delete unistr;
	return size_t(result);
}
