/*  -*- c++ -*-
    $Id: CppCFString.h,v 1.11 2003/06/22 10:56:40 phonohawk Exp $
    ---------
	
    MacOS X implementation of the SKK input method.
    Copyright (C) 2002 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#pragma once
#include <Carbon/Carbon.h>
#include <vector>
using namespace std;

// CFString̃bp[B
// CFMutableStringRefbvAstd::stringjava.lang.StringBuffer̂悤Ɉ悤ɂB

class CppCFString {
protected:
    CFMutableStringRef cf_string;
    
public:
    CppCFString();
    CppCFString(const CppCFString& cpp_cf_string_src);
    CppCFString(UniChar c); // ꕶ܂CppCFString쐬B
    CppCFString(const char* c_string);
    CppCFString(const char* c_string,CFStringEncoding encoding);
    CppCFString(CFStringRef cf_string_src);
    CppCFString(const UniChar* uni_string,const unsigned int uni_string_len);
    CppCFString(CFDataRef cfdata);
    virtual ~CppCFString();
    
    virtual CppCFString& append(const CppCFString& str);
    virtual CppCFString& append(const UniChar c);
    virtual CppCFString& insert(unsigned offset,UniChar c);
    virtual CppCFString& insert(unsigned offset,const CppCFString& str);
    virtual CppCFString& clear();
    virtual unsigned int length() const;
    virtual CppCFString& erase(unsigned start_pos,unsigned end_pos);
    virtual CppCFString& eraseCharAt(unsigned pos);
    virtual CppCFString& eraseFirst(unsigned int n_chars);
    virtual CppCFString& eraseLast(unsigned int n_chars);
    
    // java.lang.String#substringB
    virtual CppCFString substring(unsigned int start_pos) const;
    virtual CppCFString substring(unsigned int start_pos,unsigned int end_pos) const;
    virtual CppCFString trim() const;
    
    virtual int indexOf(UniChar c,unsigned int search_start_pos = 0) const;
    virtual bool startsWith(const CppCFString& str,int search_start_pos = 0) const;
    virtual bool endsWith(const CppCFString& str,int search_start_pos = -1) const;
    
    virtual CppCFString& replace(const CppCFString& from,const CppCFString& to);
    virtual CppCFString replaceClone(const CppCFString& from,const CppCFString& to) const;
    
    virtual UniChar operator[] (unsigned int index) const; // Q
    //virtual UniChar& operator[] (unsigned int index); // 
    virtual CppCFString& operator= (const CppCFString& str);
    virtual CppCFString& operator+= (const CppCFString& str);
    virtual CppCFString& operator+= (const UniChar c);
    
    virtual CFStringRef getString() const;
    virtual UniChar* toUniCharPtr() const;
    virtual CFDataRef toCFData() const;
    virtual char* toCString(CFStringEncoding encoding = kCFStringEncodingMacRomanLatin1) const;
    virtual CppCFString clone() const;
    
    virtual vector<CppCFString> split(UniChar delim) const;
};

CppCFString operator+ (const CppCFString& l,const CppCFString& r);
CppCFString operator+ (const CppCFString& l,const UniChar r);
CppCFString operator+ (const UniChar l,const CppCFString& r);

bool operator== (const CppCFString& l,const CppCFString& r);
bool operator!= (const CppCFString& l,const CppCFString& r);
bool operator< (const CppCFString& l,const CppCFString& r);
bool operator> (const CppCFString& l,const CppCFString& r);
bool operator<= (const CppCFString& l,const CppCFString& r);
bool operator<= (const CppCFString& l,const CppCFString& r);

CppCFString join(UniChar delim,const vector<CppCFString>& vec);


class hash_CppCFString {
public:
    size_t operator() (const CppCFString& str) const;
};
