/*
	$Id: KanaTreeElement.cpp,v 1.4 2002/11/29 12:29:10 phonohawk Exp $
	---------
	
    MacOS X implementation of the SKK input method.
    Copyright (C) 2002 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "KanaTreeElement.h"

KanaTreeElement::KanaTreeElement(const CppCFString& content_hiragana,
				const CppCFString& content_zenkata,
				const CppCFString& content_hankata,
				const UniChar label,
				const UniChar next_state)
		:content_hiragana(content_hiragana),
		 content_zenkata(content_zenkata),
		 content_hankata(content_hankata),
		 label(label),
		 next_state(next_state)
{
	
}

KanaTreeElement::KanaTreeElement(const KanaTreeElement& copy_source)
	:content_hiragana(copy_source.getContentAsHiragana()),
	 content_zenkata(copy_source.getContentAsZenKata()),
	 content_hankata(copy_source.getContentAsHanKata()),
	 label(copy_source.getLabel()),
	 next_state(copy_source.getNextState()),
	 childs(copy_source.getChilds())
{

}

KanaTreeElement::~KanaTreeElement()
{

}
		
KanaTreeElement& KanaTreeElement::addChild(const KanaTreeElement& child)
{
	childs.push_back(child);
	return *this;
}

KanaTreeElement& KanaTreeElement::trace(UniChar label)
{
	for (vector<KanaTreeElement>::iterator child = childs.begin();child != childs.end();child++)
	{
		if (child->getLabel() == label) // t
		{
			return *child;
		}
	}
	// tȂ
	throw ChildNotFoundException();
}

const CppCFString& KanaTreeElement::getContentAsHiragana() const
{
	return content_hiragana;
}

const CppCFString& KanaTreeElement::getContentAsZenKata() const
{
	return content_zenkata;
}

const CppCFString& KanaTreeElement::getContentAsHanKata() const
{
	return content_hankata;
}

UniChar KanaTreeElement::getLabel() const
{
	return label;
}

UniChar KanaTreeElement::getNextState() const
{
	return next_state;
}

bool KanaTreeElement::hasAnyChilds() const
{
	return childs.size() > 0;
}

const vector<KanaTreeElement>& KanaTreeElement::getChilds() const
{
	return childs;
}