/*
	$Id: SKKDictionary.cpp,v 1.14 2003/11/06 02:25:40 phonohawk Exp $
	---------
	
    MacOS X implementation of the SKK input method.
    Copyright (C) 2002 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <fstream>
#include <unistd.h>
#include <pthread.h>
#include "SKKDictionary.h"
#include "StringTokenizer.h"
using namespace util;

static void parserThreadRoutine(SKKDictionary* dic); // 񃁃o

SKKDictionary::SKKDictionary(const string& fpath,bool read_only)
	: read_only(read_only),
	 fpath(fpath)
{
	ifstream ifs(fpath.c_str(),ios::in);
	
	if (!ifs.fail())
	{
		load(ifs);
		ifs.close();
		
		// p[TXbhN
		pthread_t parser;
		pthread_create(&parser,NULL,(void*(*)(void*))parserThreadRoutine,this);
		pthread_detach(parser);
	}
}

SKKDictionary::~SKKDictionary()
{
	
}

void SKKDictionary::changeDictionaryFile(const string& fpath)
{
	okuri_ari_table.clear();
	okuri_nasi_table.clear();
	okuri_ari_unparsed.clear();
	okuri_nasi_unparsed.clear();
	
	ifstream ifs(fpath.c_str(),ios::in);
	if (!ifs.fail())
	{
		load(ifs);
		ifs.close();
	}
	parseAll();
}

void SKKDictionary::load(istream& is)
{	
	string str_line;
	while (!is.eof())
	{
		getline(is,str_line,'\n');
		if (str_line == ";; okuri-ari entries.") break;
	}
	while (!is.eof())
	{
		getline(is,str_line,'\n');
		if (str_line.length() == 0) continue;
		if (str_line == ";; okuri-nasi entries.") break;
		if (str_line[0] == ';') continue;
		// LGg
		//CppCFString line(str_line.c_str(),kCFStringEncodingEUC_JP); // EUC-JP
		
		okuri_ari_unparsed.push_back(str_line);
	}
	while (!is.eof())
	{
		getline(is,str_line,'\n');
		if (str_line.length() == 0) continue;
		if (str_line[0] == ';') continue;
		// 薳Gg
		
		okuri_nasi_unparsed.push_back(str_line);
	}
}

void SKKDictionary::save(bool& terminate_flag)
{
	if (read_only) return;
	
	string dic_tmp_file = fpath + ".tmp";
	ofstream ofs(dic_tmp_file.c_str(),ios::out);
	
	try {
		save(ofs,terminate_flag);
	}
	catch (int e) {
		ofs.close();
		unlink(dic_tmp_file.c_str()); // .tmpt@C폜
		throw e; // OĂѓ
	}
	
	ofs.flush();
	ofs.close();
	rename(dic_tmp_file.c_str(),fpath.c_str());
}

void SKKDictionary::save(ostream& ofs,bool& terminate_flag)
{
	// terminate_flagtrueɂȂint(-1)throwB
	// ̎_łmapRs[B
	//map<CppCFString,vector<OkuriganaEntry>*> okuri_ari_table_copy = okuri_ari_table;
		
	ofs << ";; okuri-ari entries." << endl;
	if (okuri_ari_table.size() > 0)
	{
		for (map<CppCFString,vector<OkuriganaEntry>*>::const_iterator entry = okuri_ari_table.begin();entry != okuri_ari_table.end();entry++)
		{
			if (entry->second == NULL) continue;
			if (entry->second->size() == 0) continue;
			if (terminate_flag) // ~w
			{
				throw int(-1);
			}
			
			CppCFString line;
			line.append(entry->first).append(" /");
			
			// ŏ̓ChJ[huB
			for (vector<OkuriganaEntry>::const_iterator okuri = entry->second->begin();okuri != entry->second->end();okuri++)
			{
				if (okuri->isWild()) // ChJ[h
				{
					line.append(join('/',okuri->getCandidates())).append('/');
					break;
				}
			}
			
			// ɃChJ[hȊOuB
			for (vector<OkuriganaEntry>::const_iterator okuri = entry->second->begin();okuri != entry->second->end();okuri++)
			{
				if (okuri->isWild()) continue;
				
				// k /u/[/u/]/
				line.append('[').append(okuri->getKana()).append('/');
				line.append(join('/',okuri->getCandidates())).append("/]/");
			}
			
			// EUC-JPɕϊďށB
			ofs << line.toCString(kCFStringEncodingEUC_JP) << endl;
		}
	}
	
	
	// ̎_łmapRs[B
	//map<CppCFString,vector<CppCFString>*> okuri_nasi_table_copy = okuri_nasi_table;
	
	ofs << ";; okuri-nasi entries." << endl;
	if (okuri_nasi_table.size() > 0)
	{
		for (map<CppCFString,vector<CppCFString>*>::const_iterator entry = okuri_nasi_table.begin();entry != okuri_nasi_table.end();entry++)
		{
			if (entry->second == NULL) continue;
			if (entry->second->size() == 0) continue;
			if (terminate_flag) // ~w
			{
				throw int(-1);
			}
			
			CppCFString line;
			line.append(entry->first).append(" /");
			
			// SĂ̍ڂuB
			line.append(join('/',*(entry->second))).append('/');
			
			// EUC-JPɕϊďށB
			ofs << line.toCString(kCFStringEncodingEUC_JP) << endl;
		}
	}
}

vector<CppCFString> SKKDictionary::findWithoutOkuri(const CppCFString& query)
{
	// ݂Ȃ΋vectorԂB
	vector<CppCFString> *candidates = okuri_nasi_table[query];
	if (candidates == NULL) // Gg݂̑ȂA܂p[XĂȂB
	{
		for (unsigned i = 0;i < okuri_nasi_unparsed.size();i++)
		{
			CppCFString line(okuri_nasi_unparsed[i].c_str(),kCFStringEncodingEUC_JP); // EUC-JP
			if (line.startsWith(query))
			{
				CppCFString index;
				vector<CppCFString> entries;
				parseOkuriNasiEntry(line,index,entries);
				
				return entries;
			}
		}
		
		vector<CppCFString> empty;
		return empty;
	}
	else
	{
		return *candidates;
	}
}

vector<OkuriganaEntry> SKKDictionary::findWithOkuri(const CppCFString& root)
{
	// ݂Ȃ΋vectorԂB
	vector<OkuriganaEntry> *candidates = okuri_ari_table[root];
	if (candidates == NULL) // Gg݂̑ȂA܂p[XĂȂB
	{
		for (unsigned i = 0;i < okuri_ari_unparsed.size();i++)
		{
			CppCFString line(okuri_ari_unparsed[i].c_str(),kCFStringEncodingEUC_JP); // EUC-JP
			if (line.startsWith(root))
			{
				CppCFString index;
				vector<OkuriganaEntry> entries;
				parseOkuriAriEntry(line,index,entries);
				
				return entries;
			}
		}
		
		vector<OkuriganaEntry> empty;
		return empty;
	}
	else
	{
		return *candidates;
	}
}

vector<CppCFString> SKKDictionary::findCompletions(const CppCFString& query)
{
	vector<CppCFString> result;
	const unsigned query_len = query.length();
	
	// ݂Ȃ΋vectorԂB
	for (map<CppCFString,vector<CppCFString>*>::const_iterator ite = okuri_nasi_table.begin();ite != okuri_nasi_table.end();ite++)
	{
		if (ite->first.length() > query_len && ite->first.startsWith(query))
		{
			// ⊮̏𖞂B
			result.push_back(ite->first);
		}
	}
	return result;
}

const string& SKKDictionary::getFilePath() const
{
	return fpath;
}

void SKKDictionary::parseAll()
{
	// @̃\bh́ALGgQƑ薳GgQ
	// 납p[XĂAunparsedvector͏čsB
	// OłȂŔAO̍ڂɌɂ鍀ڂ
	// Oɂ炳Ă܂mȂׂłB炷ɂ͗]vȎԂB
	// @ȂÃ\bhload()OɌĂł͂ȂȂB
	
	// LGg
	for (vector<string>::iterator ite = okuri_ari_unparsed.end()-1;okuri_ari_unparsed.size() > 0;ite--)
	{
		CppCFString line(ite->c_str(),kCFStringEncodingEUC_JP); // EUC-JP
		
		CppCFString index;
		vector<OkuriganaEntry> entries;
		parseOkuriAriEntry(line,index,entries);
		
		vector<OkuriganaEntry>* candidates = okuri_ari_table[index];
		if (candidates == NULL)
		{
			candidates = new vector<OkuriganaEntry>;
			okuri_ari_table[index] = candidates;
		}
		candidates->insert(candidates->end(),entries.begin(),entries.end());
		
		okuri_ari_unparsed.erase(ite);
	}
	
	// 薳Gg
	for (vector<string>::iterator ite = okuri_nasi_unparsed.end()-1;okuri_nasi_unparsed.size() > 0;ite--)
	{
		CppCFString line(ite->c_str(),kCFStringEncodingEUC_JP); // EUC-JP
		
		CppCFString index;
		vector<CppCFString> entries;
		parseOkuriNasiEntry(line,index,entries);
		
		vector<CppCFString>* candidates = okuri_nasi_table[index];
		if (candidates == NULL)
		{
			candidates = new vector<CppCFString>;
			okuri_nasi_table[index] = candidates;
		}
		candidates->insert(candidates->end(),entries.begin(),entries.end());
		
		okuri_nasi_unparsed.erase(ite);
	}
	
	cout << fpath << endl;
	cout << "  okuri-ari: " << okuri_ari_table.size() << " entries." << endl;
	cout << "  okuri-nasi: " << okuri_nasi_table.size() << " entries." << endl;
	cout << endl;
}

void SKKDictionary::registerWord(const CppCFString& index,const CppCFString& okuri,const CppCFString& kanji)
{
	// 艼LB
	vector<OkuriganaEntry>* candidates = okuri_ari_table[index];
	if (candidates == NULL)
	{
		candidates = new vector<OkuriganaEntry>;
		okuri_ari_table[index] = candidates;
		
		// ₪ꍇ̓ChJ[hɊďIB
		OkuriganaEntry entry;
		entry.add(kanji);
		
		candidates->push_back(entry);
	}
	else
	{
		// ChJ[h̐擪ɂ̊ǉB
		for (vector<OkuriganaEntry>::iterator ite = candidates->begin();ite != candidates->end();ite++)
		{
			if (ite->isWild())
			{
				ite->insert(kanji,0);
				ite->removeRedundantItems();
				break;
			}
		}
		
		// 艼OkuriganaEntry΁AGgĊďIB
		// ɂ΂̃Gg̐擪ɊǉAvector̐擪ɒuB
		bool found = false;
		for (vector<OkuriganaEntry>::iterator ite = candidates->begin();ite != candidates->end();ite++)
		{
			if (ite->getKana() == okuri)
			{
				found = true;
				ite->insert(kanji,0);
				ite->removeRedundantItems(); // d폜
				
				//candidates.swap(candidates.begin(),ite); // 擪̍ڂƓւB
				OkuriganaEntry entry = *ite;
				candidates->erase(ite);
				candidates->insert(candidates->begin(),entry);
				break;
			}
		}
		if (!found)
		{
			OkuriganaEntry entry(okuri);
			entry.add(kanji);
			candidates->insert(candidates->begin(),entry);
		}
	}
}

void SKKDictionary::registerWord(const CppCFString& index,const CppCFString& kanji)
{
	vector<CppCFString>* candidates = okuri_nasi_table[index];
	if (candidates == NULL)
	{
		candidates = new vector<CppCFString>;
		okuri_nasi_table[index] = candidates;
	}
	
	// ɂ̊o^Ă΁AB
	for (vector<CppCFString>::iterator ite = candidates->begin();ite != candidates->end();ite++)
	{
		if (*ite == kanji)
		{
			candidates->erase(ite);
			break;
		}
	}
	
	// 擪ɓB
	candidates->insert(candidates->begin(),kanji);
}

void SKKDictionary::removeWord(const CppCFString& index,const CppCFString& okuri,const CppCFString& kanji)
{
	// 艼
	vector<OkuriganaEntry>* candidates = okuri_ari_table[index];
	if (candidates == NULL) return;
	
	// SĂOkuriganaEntryAkanjił̂TĂB
	for (vector<OkuriganaEntry>::iterator entry = candidates->begin();entry != candidates->end();entry++)
	{
		unsigned n_contents = entry->size();
		for (unsigned i = 0;i < n_contents;i++)
		{
			if ((*entry)[i] == kanji)
			{
				entry->remove(i);
				i--;
				n_contents--;
			}
		}
	}
}

void SKKDictionary::removeWord(const CppCFString& index,const CppCFString& kanji)
{
	vector<CppCFString>* candidates = okuri_nasi_table[index];
	if (candidates == NULL) return;
	
	unsigned size = candidates->size();
	for (unsigned i = 0;i < size;i++)
	{
		if ((*candidates)[i] == kanji)
		{
			candidates->erase(candidates->begin()+i);
			i--;
			size--;
		}
	}
}

// 񃁃o֐Q
void parseOkuriAriEntry(const CppCFString& line,CppCFString& index,vector<OkuriganaEntry>& entries)
{
	// s /o/[/o/]/[/o/]/[/o/]/
	int pos_space = line.indexOf(' ');
	if (pos_space == -1) return; // ُȃGg
	
	index = line.substring(0,pos_space);
	CppCFString candidates_str(line.substring(pos_space+1,line.length()));
	
	// [Ŏn܂Gg܂ł̓ChJ[hB
	OkuriganaEntry wildcard(CppCFString(""));
	const int len = candidates_str.length();
	int pos = 1;
	while (pos < len)
	{
		// Gg̎n܂
		if (candidates_str[pos] != '[') // ChJ[ȟ
		{
			int blockend = candidates_str.indexOf('/',pos);
			wildcard.add(candidates_str.substring(pos,blockend));
			
			pos = blockend+1;
		}
		else // ChJ[ȟ
		{
			int entry_end_pos = candidates_str.indexOf(']',pos);
			OkuriganaEntry entry(CppCFString(""));
			bool this_is_the_first_element = true; // []̍ŏ̍ڂǂBŏ̍ڂoƂȂB
			pos++; // [̎B
			while (pos < entry_end_pos)
			{
				int blockend = candidates_str.indexOf('/',pos);
				if (this_is_the_first_element)
				{
					entry.setKana(candidates_str.substring(pos,blockend));
					this_is_the_first_element = false;
				}
				else
				{
					entry.add(candidates_str.substring(pos,blockend));
				}
				pos = blockend+1;
			}
			entries.push_back(entry);
			
			pos = entry_end_pos+2;
		}
	}
	entries.push_back(wildcard);
}

void parseOkuriNasiEntry(const CppCFString& line,CppCFString& index,vector<CppCFString>& entries)
{
	//  /w/Ȋw/
	int pos_space = line.indexOf(' ');
	if (pos_space == -1) return; // ُȃGg
	
	index = line.substring(0,pos_space);
	CppCFString candidates_str = line.substring(pos_space+1);
	
	entries = candidates_str.split('/');
}

static void parserThreadRoutine(SKKDictionary* dic)
{
	dic->parseAll();
}
