/*
	$Id: SKKServer.cpp,v 1.16 2003/06/22 10:56:40 phonohawk Exp $
	---------
	
    MacOS X implementation of the SKK input method.
    Copyright (C) 2002 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "SKKServer.h"
#include "StringUtils.h"
using namespace util;

#if 0
#	define D_PRINTF printf
#else
#	define D_PRINTF if(0)printf
#endif

static SKKServer* _shared_instance = NULL;

SKKServer& SKKServer::sharedServer()
{
	return *_shared_instance;
}

SKKServer& SKKServer::sharedServer(const string& base_dic_file,const string& sub_dic_file,const string& user_dic_file,const vector<CppCFString>& kotoeri_dic_file)
{
	if (_shared_instance == NULL)
	{
		_shared_instance = new SKKServer(base_dic_file,sub_dic_file,user_dic_file,kotoeri_dic_file);
	}
	return *_shared_instance;
}

SKKServer::SKKServer(const string& base_dic_file,const string& sub_dic_file,const string& user_dic_file,const vector<CppCFString>& kotoeri_dic_file)
	:base_dic(base_dic_file,true) , sub_dic(sub_dic_file,true) , user_dic(user_dic_file,false) , kotoeri_dic(kotoeri_dic_file)
{
	all_dics.push_back(&user_dic);
	all_dics.push_back(&sub_dic);
	all_dics.push_back(&base_dic);
    all_dics.push_back(&kotoeri_dic);
}

SKKServer::~SKKServer()
{

}

void SKKServer::parseAllDictionaries()
{
	user_dic.parseAll();
	sub_dic.parseAll();
	base_dic.parseAll();
}

SKKDictionary& SKKServer::getBaseDic()
{
	return base_dic;
}

SKKDictionary& SKKServer::getSubDic()
{
	return sub_dic;
}

SKKDictionary& SKKServer::getUserDic()
{
	return user_dic;
}

DMDictionary& SKKServer::getKotoeriDic()
{
	return kotoeri_dic;
}

CppCFString SKKServer::skkserv_style_search(const CppCFString& query) {
  // skkservX^Cł̞sӁB
  // NǦ`́AuƂvusv̂₤Ȃ́Bɋ󔒕͕tȂB
  // vČ`́Au/擪/퓬/K/v̂₤ɁA熏̌`̂܂܁B΋󕶎B
  if (query.length() == 0) {
    return CppCFString();
  }

  // ͗L邩H
  if (query[0] > 0xff &&
      query[query.length()-1] >= 'a' && query[query.length()-1] <= 'z') {
    // LB

    map<CppCFString,OkuriganaEntry> cands; // OkuriganaEntry.kana => OkuriganaEntry
    
    for (vector<Dictionary*>::const_iterator dic = all_dics.begin();
	 dic != all_dics.end(); dic++) {
      
      vector<OkuriganaEntry> vec_okuri = (*dic)->findWithOkuri(query);
      if (vec_okuri.size() == 0) {
	continue;
      }

      for (vector<OkuriganaEntry>::iterator e = vec_okuri.begin();
	   e != vec_okuri.end(); e++) {
	  
	  OkuriganaEntry& entry = cands[e->getKana()];
	  entry.setKana(e->getKana());
	  entry.getCandidates().insert(
	      entry.getCandidates().begin(),
	      e->getCandidates().begin(),
	      e->getCandidates().end());
      }
    }
    
    // SẴGgɂāAdĂ̂폜ȂAvCgĂB
    CppCFString reply;
    for (map<CppCFString,OkuriganaEntry>::iterator e = cands.begin();
	 e != cands.end(); e++) {
	OkuriganaEntry& entry = e->second;
	entry.removeRedundantItems();

	if (entry.isWild()) {
	    reply.append(join('/',entry.getCandidates())).append('/');
	}
	else {
	    reply.append('[').append(entry.getKana()).append('/').
		append(join('/',entry.getCandidates())).append("/]/");
	}
    }
    if (reply.length() > 0) {
	reply.insert(0,'/');
    }
    return reply;
  }
  else {
      // B
      vector<CppCFString> cands;
      
      for (vector<Dictionary*>::const_iterator ite = all_dics.begin();
	   ite != all_dics.end(); ite++) {
	  
	  vector<CppCFString> cand_for_one = (*ite)->findWithoutOkuri(query);
	  if (cand_for_one.size() > 0) {
	      cands.insert(cands.end(),cand_for_one.begin(),cand_for_one.end());
	  }
      }

      // dĂ̂폜B
      removeRedundantItems(cands);

      if (cands.size() > 0) {
	  return CppCFString('/').append(join('/',cands)).append('/');
      }
      else {
	  return CppCFString();
      }
  }
}

CppCFString SKKServer::search(const CppCFString& query)
{
	CppCFString space(" ");
	CppCFString nbsp("[20]");
	
	bool has_okuri = query[0] == '+';
	
	CppCFString query_str(query.substring(1)); // 擪+,-B
	
	//cout << "QUERY: " << query_str << endl;
	D_PRINTF("QUERY: %s\n",query_str.toCString(kCFStringEncodingEUC_JP));
	
	if (has_okuri)
	{
		int pos_space = query_str.indexOf(' ');
		CppCFString root = query_str.substring(0,pos_space).replace(nbsp,space); // u+Ȃr vȂuȂrv
		CppCFString okuri = query_str.substring(pos_space+1).replace(nbsp,space); // u+Ȃr vȂuv
		
		vector<CppCFString> cand_strictly_matched;
		vector<CppCFString> cand_unstrictly_matched;
		for (vector<Dictionary*>::const_iterator dic = all_dics.begin();dic != all_dics.end();dic++)
		{
			vector<OkuriganaEntry> vec_okuri = (*dic)->findWithOkuri(root);
			if (vec_okuri.size() == 0) continue;
			
			for (vector<OkuriganaEntry>::const_iterator okuri_entry = vec_okuri.begin();
				 okuri_entry != vec_okuri.end();
				 okuri_entry++)
			{
				// OkuriganaEntry̌ookurivĂcand_strictly_matchedɓA
				// vĂȂcand_unstrictly_matchedɓB
				if (okuri_entry->getKana() == okuri)
				{
					cand_strictly_matched.insert(
					    cand_strictly_matched.end(),
					    okuri_entry->getCandidates().begin(),
					    okuri_entry->getCandidates().end());
				}
				else
				{
					cand_unstrictly_matched.insert(cand_unstrictly_matched.end(),
												   okuri_entry->getCandidates().begin(),
												   okuri_entry->getCandidates().end());
				}
			}
		}
		vector<CppCFString> candidates = cand_strictly_matched;
		candidates.insert(candidates.end(),cand_unstrictly_matched.begin(),cand_unstrictly_matched.end());
		
		// candidates̏d`FbNBdĂɂ̂폜B
		removeRedundantItems(candidates);
		
		// ̎_candidatesɓĂ̂͊̕Ȃ̂ŁA
		// SĂ̗vfɑ艼tBɃXy[X[20]ɕϊB
		for (vector<CppCFString>::iterator ite = candidates.begin();ite != candidates.end();ite++)
		{
			ite->append(okuri).replace(space,nbsp);
		}
		
		D_PRINTF("REPLY: %s\n",join(' ',candidates).toCString(kCFStringEncodingEUC_JP));
		return join(' ',candidates);
	}
	else
	{
		vector<CppCFString> candidates;
		
		for (vector<Dictionary*>::const_iterator ite = all_dics.begin();ite != all_dics.end();ite++)
		{
			vector<CppCFString> cand_for_one = (*ite)->findWithoutOkuri(query_str);
			//if (cand_for_one.size() > 0)
			//	cout << "dic: " << join(' ',cand_for_one).toCString(kCFStringEncodingEUC_JP) << endl;
			if (cand_for_one.size() > 0)
				candidates.insert(candidates.end(),cand_for_one.begin(),cand_for_one.end());
		}
		
		// candidates̏d`FbNBdĂɂ̂폜B
		removeRedundantItems(candidates);
		
		// candidates̑SĂ̍ڂ̃Xy[X[20]ɕϊB
		for (vector<CppCFString>::iterator ite = candidates.begin();ite != candidates.end();ite++)
		{
			ite->replace(space,nbsp);
		}
		
		D_PRINTF("REPLY: %s\n",join(' ',candidates).toCString(kCFStringEncodingEUC_JP));
		return join(' ',candidates);
	}
}

void SKKServer::registerToUserDic(const CppCFString& query)
{
	CppCFString space(" ");
	CppCFString nbsp("[20]");
	
	D_PRINTF("REGISTER: %s\n",query.toCString(kCFStringEncodingEUC_JP));
	
	if (query[0] == '+') // 艼Bu+r  v
	{
		int pos_first_space = query.indexOf(' ');
		int pos_second_space = query.indexOf(' ',pos_first_space+1);
		
		CppCFString index(query.substring(1,pos_first_space)); index.replace(nbsp,space); // r
		CppCFString okuri(query.substring(pos_first_space+1,pos_second_space)); okuri.replace(nbsp,space); // 
		CppCFString kanji(query.substring(pos_second_space+1)); kanji.replace(nbsp,space); // 
		
		if (index.length() == 0 || okuri.length() == 0 || kanji.length() == 0)
		{
			cout << "Invalid registration received; index:" << index.length()
				 << " okuri:" << okuri.length() << " kanji:" << kanji.length() << " (len)" << endl;
		}
		else
		{
			user_dic.registerWord(index,okuri,kanji);
		}
	}
	else // 艼Bu- v
	{
		int pos_space = query.indexOf(' ');
		CppCFString index(query.substring(1,pos_space)); index.replace(nbsp,space);
		CppCFString kanji(query.substring(pos_space+1)); kanji.replace(nbsp,space);
		
		if (index.length() == 0 || kanji.length() == 0)
		{
			cout << "Invalid registration received; index:" << index.length()
				 << " kanji:" << kanji.length() << " (len)" << endl;
		}
		else
		{
			user_dic.registerWord(index,kanji);
		}
	}
}

void SKKServer::removeFromUserDic(const CppCFString& query)
{
	CppCFString space(" ");
	CppCFString nbsp("[20]");
	
	D_PRINTF("REMOVE: %s\n",query.toCString(kCFStringEncodingEUC_JP));
	
	if (query[0] == '+') // 艼Bu+r  v
	{
		int pos_first_space = query.indexOf(' ');
		int pos_second_space = query.indexOf(' ',pos_first_space+1);
		
		CppCFString index(query.substring(1,pos_first_space)); index.replace(nbsp,space); // r
		CppCFString okuri(query.substring(pos_first_space+1,pos_second_space)); okuri.replace(nbsp,space); // 
		CppCFString kanji(query.substring(pos_second_space+1)); kanji.replace(nbsp,space); // 
		
		if (index.length() == 0 || okuri.length() == 0 || kanji.length() == 0)
		{
			cout << "Invalid removal received; index:" << index.length()
				 << " okuri:" << okuri.length() << " kanji:" << kanji.length() << " (len)" << endl;
		}
		else
		{
			user_dic.removeWord(index,okuri,kanji);
		}
	}
	else // 艼Bu- v
	{
		int pos_space = query.indexOf(' ');
		CppCFString index(query.substring(1,pos_space)); index.replace(nbsp,space);
		CppCFString kanji(query.substring(pos_space+1)); kanji.replace(nbsp,space);
		
		if (index.length() == 0 || kanji.length() == 0)
		{
			cout << "Invalid removal received; index:" << index.length()
				 << " kanji:" << kanji.length() << " (len)" << endl;
		}
		else
		{
			user_dic.removeWord(index,kanji);
		}
	}
}

CppCFString SKKServer::searchCompletions(const CppCFString& query)
{
	CppCFString space(" ");
	CppCFString nbsp("[20]");
	
	D_PRINTF("COMPLETE: %s\n",query.toCString(kCFStringEncodingEUC_JP));
	
	// ܂̓[U[AɃTu猟B
	CppCFString head = query.replaceClone(nbsp,space);
	vector<CppCFString> user_comp = user_dic.findCompletions(head);
	vector<CppCFString> sub_comp = user_dic.findCompletions(head);
	
	vector<CppCFString> result;
	result.insert(result.end(),user_comp.begin(),user_comp.end());
	result.insert(result.end(),sub_comp.begin(),sub_comp.end());
	
	removeRedundantItems(result);
	for (vector<CppCFString>::iterator ite = result.begin();ite != result.end();ite++)
	{
		ite->replace(space,nbsp);
	}
	
	D_PRINTF("REPLY: %s\n",join(' ',result).toCString(kCFStringEncodingEUC_JP));
	return join(' ',result);
}

void removeRedundantItems(vector<CppCFString>& candidates)
{
	// ̏candidates̐قǎԂ܂B
	// dȂ̂낤ǁB
	int size = candidates.size();
	for (int i = 0;i < size;i++)
	{
		CppCFString& cand_i = candidates[i];
		
		bool is_redundant = false;
		for (int j = 0;j < i;j++)
		{
			if (cand_i == candidates[j])
			{
				is_redundant = true;
				break;
			}
		}
		
		if (is_redundant)
		{
			candidates.erase(candidates.begin()+i);
			i--; // ̂ŁǍɑڂB
			size--; // ̂ŁB
		}
	}
}
