/*
 *  $Id: SkkConfig.cpp,v 1.5 2003/05/14 12:03:14 phonohawk Exp $
 *
 *  KeymapConfig.cpp
 *  AquaSKK
 *
 *  Created by Ђ낦` on Fri Nov 15 2002.
 *  Copyright (c) 2002 Hre-I. All rights reserved.
 *
 */
#include <fstream.h>
#include <sys/types.h>
#include <pwd.h>
#include <unistd.h>
#include "InputMode.h"
#include "SkkConfig.h"


/**
 *
 */
enum SkkValCode {
	SKK_UNKNOWN,
	SKK_TOGGLE_KANA,
	SKK_TOGGLE_KATAKANA,
	SKK_KANA_MODE,	
	SKK_LATIN_MODE,
	SKK_JISX0208_LATIN_MODE,
	SKK_ABBREV_MODE,
	SKK_KAKUTEI,
	SKK_PREV_KOUHO,
	SKK_NEXT_KOUHO,
	SKK_ABBREV_PREV,
	SKK_ABBREV_NEXT,
	SKK_PURGE_FROM_JISYO,
	SKK_CANCEL,
	SKK_FIX_N,
	SKK_SET_HENKAN_POINT,
	SKK_PASTE,
        SKK_USE_EISUU_TO_SET_HENKAN_POINT,
        SKK_USE_KANA_TO_SET_HENKAN_POINT,
};


struct SkkVariable {
	SkkValCode	code;
	const char*	str;
} variables[] = {
	{ SKK_TOGGLE_KANA,		"toggle-kana-key" },
	{ SKK_TOGGLE_KATAKANA,		"toggle-katakana-key" },
	{ SKK_KANA_MODE,		"kana-mode-key" },
	{ SKK_LATIN_MODE,		"latin-mode-key" },
	{ SKK_JISX0208_LATIN_MODE,	"jisx0208-latin-mode-key" },
	{ SKK_ABBREV_MODE,		"abbrev-mode-key" },
	{ SKK_KAKUTEI,			"kakutei-key" },
	{ SKK_PREV_KOUHO,		"prev-kouho-key" },
	{ SKK_NEXT_KOUHO,		"next-kouho-key" },
	{ SKK_ABBREV_PREV,		"abbrev-prev-key" },
	{ SKK_ABBREV_NEXT,		"abbrev-next-key" },
	{ SKK_PURGE_FROM_JISYO,		"purge-from-jisyo-key" },
	{ SKK_CANCEL,			"cancel-key" },
	{ SKK_FIX_N,			"fix-n" },
	{ SKK_SET_HENKAN_POINT,			"set-henkan-point-key" },
	{ SKK_PASTE,			"paste-key" },
	{ SKK_USE_EISUU_TO_SET_HENKAN_POINT,			"use-eisuu-to-set-henkan-point" },
	{ SKK_USE_KANA_TO_SET_HENKAN_POINT,			"use-kana-to-set-henkan-point" },
};
#define	VARIABLE_NUM	(sizeof(variables)/sizeof(variables[0]))


/**
 * SkkConfig ̗B̃CX^X
 */
SkkConfig* SkkConfig::config_ = NULL;

/**
 *
 */
static SkkValCode	parseVariable(string& str);
static bool		parseBoolean(string& str);
static uchar		parseKeycode(string& str);


/**
 * Constructor for TUT Code mode
 */
SkkConfig::SkkConfig()
{
	fixN_ = true;
	useEisuuToSetHenkanPoint_ = false;	// defaultfalse
	useKanaToSetHenkanPoint_ = false;	// defaultfalse

	// ftHgL[}bv
	kanaModeKey_ 		= 'J' - 0x40;
	toggleKanaKey_		= 'q';
	toggleKatakanaKey_	= 'Q' - 0x40;
	latinModeKey_ 		= 'l';
	jisx0208LatinModeKey_ 	= 'L';
	abbrevModeKey_		= '/';
	cancelKey_		= 'G' - 0x40;
	kakuteiKey_		= 'J' - 0x40;
	nextKouhoKey_		= ' ';
	prevKouhoKey_		= 'x';
	purgeFromJisyoKey_	= 'X';
	abbrevPrevKey_		= ',';
	abbrevNextKey_		= '.';
	setHenkanPointKey_	= 'Q';
	pasteKey_			= 'Y' - 0x40;
	
	string config_file = home() + "/Library/AquaSKK/config";
	ifstream config_fs(config_file.c_str(), ios::in);
	if (!config_fs.fail()) {
		loadConfig(config_fs);
		config_fs.close();
	}
	
	okuriHeadTable_ = NULL;
	string okuri_file = home() + "/Library/AquaSKK/okuri-list";
	ifstream okuri_fs(okuri_file.c_str(), ios::in);
	if (!okuri_fs.fail()) {
		loadOkuriHeadTable(okuri_fs);
		okuri_fs.close();
	}
	
	for (int i = 0; i < 256; i++)
		mappedKey_[i] = false;
	mappedKey_[kanaModeKey_] = true;
	mappedKey_[latinModeKey_] = true;
	mappedKey_[jisx0208LatinModeKey_] = true;
	mappedKey_[toggleKanaKey_] = true;
	mappedKey_[toggleKatakanaKey_] = true;
	mappedKey_[kakuteiKey_] = true;
	mappedKey_[abbrevModeKey_] = true;
	mappedKey_[prevKouhoKey_] = true;
	mappedKey_[nextKouhoKey_] = true;
	mappedKey_[purgeFromJisyoKey_] = true;
	mappedKey_[cancelKey_] = true;
	mappedKey_[abbrevPrevKey_] = true;
	mappedKey_[abbrevNextKey_] = true;
	mappedKey_[setHenkanPointKey_] = true;
	mappedKey_[pasteKey_] = true;
}


/**
 *
 */
SkkConfig::~SkkConfig()
{
	if (okuriHeadTable_)
		delete okuriHeadTable_;
}


/**
 * config t@C̓ǂݍ
 */
void SkkConfig::loadConfig(istream& is)
{
	string str_line;
	SkkValCode valcode;
	while (!is.eof()) {
		getline(is, str_line, '\n');
		if (str_line.length() == 0) 
			continue;
		if (str_line[0] == '#') 
			continue;

		valcode = parseVariable(str_line);
		if  (valcode == SKK_FIX_N)
			fixN_ = parseBoolean(str_line);
		else if (valcode == SKK_KANA_MODE) 
			kanaModeKey_ = parseKeycode(str_line);
		else if (valcode == SKK_LATIN_MODE)
			latinModeKey_ = parseKeycode(str_line);
		else if (valcode == SKK_JISX0208_LATIN_MODE)
			jisx0208LatinModeKey_ = parseKeycode(str_line);
		else if (valcode == SKK_TOGGLE_KANA)
			toggleKanaKey_ = parseKeycode(str_line);
		else if (valcode == SKK_TOGGLE_KATAKANA)
			toggleKatakanaKey_ = parseKeycode(str_line);
		else if (valcode == SKK_ABBREV_MODE)
			abbrevModeKey_ = parseKeycode(str_line);
		else if (valcode == SKK_KAKUTEI)
			kakuteiKey_ = parseKeycode(str_line);
		else if (valcode == SKK_CANCEL)
			cancelKey_ = parseKeycode(str_line);
		else if (valcode == SKK_PREV_KOUHO)
			prevKouhoKey_ = parseKeycode(str_line);
		else if (valcode == SKK_NEXT_KOUHO)
			nextKouhoKey_ = parseKeycode(str_line);
		else if (valcode == SKK_PURGE_FROM_JISYO)
			purgeFromJisyoKey_ = parseKeycode(str_line);
		else if (valcode == SKK_SET_HENKAN_POINT)
			setHenkanPointKey_ = parseKeycode(str_line);
		else if  (valcode == SKK_USE_EISUU_TO_SET_HENKAN_POINT)
			useEisuuToSetHenkanPoint_ = parseBoolean(str_line);
		else if  (valcode == SKK_USE_KANA_TO_SET_HENKAN_POINT)
			useKanaToSetHenkanPoint_ = parseBoolean(str_line);
		else if (valcode == SKK_PASTE)
			pasteKey_ = parseKeycode(str_line);
		else
			cerr << "Unknown variable:" << str_line << endl;
	}
#ifdef DEBUG
	cerr << "kanaModeKey_" << hex << (int)kanaModeKey_ << endl;
	cerr << "latinModeKey_" << hex << (int)latinModeKey_ << endl;
	cerr << "jisx0208LatinModeKey_" << hex << (int)jisx0208LatinModeKey_ << endl;
	cerr << "toggleKanaKey_" << hex << (int)toggleKanaKey_ << endl;
	cerr << "toggleKatakanaKey_" << hex << (int)toggleKatakanaKey_ << endl;
	cerr << "kakuteiKey_" << hex << (int)kakuteiKey_ << endl;
	cerr << "abbrevModeKey_" << hex << (int)abbrevModeKey_ << endl;
	cerr << "prevKouhoKey_" << hex << (int)prevKouhoKey_ << endl;
	cerr << "nextKouhoKey_" << hex << (int)nextKouhoKey_ << endl;
	cerr << "purgeFromJisyoKey_" << hex << (int)purgeFromJisyoKey_ << endl;
	cerr << "cancelKey_" << hex << (int)cancelKey_ << endl;
	cerr << "abbrevPrevKey_" << hex << (int)abbrevPrevKey_ << endl;
	cerr << "abbrevNextKey_" << hex << (int)abbrevNextKey_ << endl;
	cerr << "pasteKey_" << hex << (int)pasteKey_ << endl;
#endif
}


#ifdef DEBUG
/**
 * @\R[h@\擾
 */
const char* getFunctionName(int code)
{
	for (unsigned i = 0; i < FUNCTION_NUM; ++i) {
		if (functions[i].code == code)
			return functions[i].str;
	}
	return "ILLEGAL FUNCTION CODE.";
}
#endif


/**
 *
 */
SkkValCode parseVariable(string& str)
{
	for (unsigned i = 0; i < VARIABLE_NUM; i++) {
		unsigned j;
		for (j = 0; j < strlen(variables[i].str); j++) {
			if (variables[i].str[j] != str[j])
				break;
		}
		if (j == strlen(variables[i].str))
			return variables[i].code;
	}
	return SKK_UNKNOWN;
}


/**
 *
 */
uchar parseKeycode(string& str) 
{
	unsigned i = 0;
	while (str[i] != '"')
		++i;
	
	if (str.compare(i,2,"\"\"") == 0) {
		return 0;
	}
	
	++i;
	if (str[i] == '\\') {
		++i;
		if (str[i] == 'C' && str[i + 1] == '-')
			return toupper(str[i + 2]) - 0x40;
		else if (str[i] == 't')
			return '\t';
		else if (str[i] == 'n')
			return '\n';
		else
			return str[i];
	} else
		return str[i];
}


/**
 *
 */
bool parseBoolean(string& str) 
{
	unsigned i = 0;
	while (!isspace(str[i]))
		++i;
	while (isspace(str[i]))
		++i;
	if (str[i] == 't')
		return true;
	return false;
}


/**
 * 艼ϊe[u̓ǂݍ
 */
void SkkConfig::loadOkuriHeadTable(istream& is)
{
	okuriHeadTable_ = new hash_map<UniChar, UniChar>();
	string str_line;
	while (!is.eof()) {
		getline(is, str_line, '\n');
		if (str_line.length() == 0) 
			continue;
		if (str_line[0] == '#') 
			continue;

		CppCFString line(str_line.c_str(), kCFStringEncodingEUC_JP);
		(*okuriHeadTable_)[line[1]] = line[0];
		//cerr << hex << line[1] << "\t" << dec << line[0] << endl;
	}
}

// [Ũz[fBNg擾
string SkkConfig::home()
{
	struct passwd* entry = getpwuid(getuid());
	string result(entry->pw_dir);
	endpwent();
	
	return result;
}
