/*
 *  $Id: SkkConfig.h,v 1.4 2003/05/14 12:03:14 phonohawk Exp $
 *
 *  KeymapConfig.h
 *  AquaSKK
 *
 *  Created by Ђ낦` on Fri Nov 15 2002.
 *  Copyright (c) 2002 Hre-I. All rights reserved.
 *
 */
#pragma once

#include <istream.h>
#include <hash_map.h>
#include "CppCFString.h"


typedef unsigned char	uchar;


/**
 * 
 */
class SkkConfig
{
	static SkkConfig*	config_;

	void clearMap();
	void loadConfig(istream&);
	void loadOkuriHeadTable(istream&);

	bool	fixN_;			// Ō n   ɕϊ邩?
	bool	useEisuuToSetHenkanPoint_;			// pL[set-henkan-pointƂĎg
	bool	useKanaToSetHenkanPoint_;			// ȃL[set-henkan-pointƂĎg

	uchar	kanaModeKey_;		// ȓ̓[h
	uchar	latinModeKey_;		// At@xbg[h
	uchar	jisx0208LatinModeKey_;	// SpAt@xbg[h
	uchar	toggleKanaKey_;		// ȃ[hƂ̃gO
	uchar	toggleKatakanaKey_;	// pJiSpJigO
	uchar	kakuteiKey_;		// mL[
	uchar	abbrevModeKey_;		// At@xbg⊮[h
	uchar	prevKouhoKey_;		// Ǒ
	uchar	nextKouhoKey_;		// ̌
	uchar	purgeFromJisyoKey_;	// 폜
	uchar	cancelKey_;		// LZL[
	uchar	abbrevPrevKey_;
	uchar	abbrevNextKey_;
	uchar	setHenkanPointKey_;	// ϊ|CgݒL[
	uchar	pasteKey_; // (o^)y[XgL[

	bool	mappedKey_[256];

	hash_map<UniChar, UniChar>*	okuriHeadTable_;


public:
	/**
	 * RXgN^CfXgN^
	 */
	SkkConfig();
	~SkkConfig();

	const bool isKanaModeKey(uchar c)		{ return kanaModeKey_ == c; }
	const bool isLatinModeKey(uchar c)		{ return latinModeKey_ == c; }
	const bool isJisx0208LatinModeKey(uchar c)	{ return jisx0208LatinModeKey_ == c; }
	const bool isToggleKanaKey(uchar c)		{ return toggleKanaKey_ == c; }
	const bool isToggleKatakanaKey(uchar c)		{ return toggleKatakanaKey_ == c; }
	const bool isKakuteiKey(uchar c)		{ return kakuteiKey_ == c; }
	const bool isAbbrevModeKey(uchar c)		{ return abbrevModeKey_ == c; }
	const bool isPrevKouhoKey(uchar c)		{ return prevKouhoKey_ == c; }
	const bool isNextKouhoKey(uchar c)		{ return nextKouhoKey_ == c; }
	const bool isPurgeFromJisyoKey(uchar c)		{ return purgeFromJisyoKey_ == c; }
	const bool isCancelKey(uchar c)			{ return cancelKey_ == c; }	
	const bool isAbbrevPrevKey(uchar c)		{ return abbrevPrevKey_ == c; }
	const bool isAbbrevNextKey(uchar c)		{ return abbrevNextKey_ == c; }
	const bool isSetHenkanPointKey(uchar c)		{ return setHenkanPointKey_ == c; }
	const bool isPasteKey(uchar c)			{ return pasteKey_ == c; }

	const uchar latinModeKey()		{ return latinModeKey_; }
	const uchar kanaModeKey()		{ return kanaModeKey_; }
	const uchar setHenkanPointKey()		{ return setHenkanPointKey_; }
	const uchar fixN()			{ return fixN_; }
	const uchar useEisuuToSetHenkanPoint()			{ return useEisuuToSetHenkanPoint_; }
	const uchar useKanaToSetHenkanPoint()			{ return useKanaToSetHenkanPoint_; }


	/**
	 *
	 */
	const bool isMappedKey(uchar c)		{ return mappedKey_[c]; }

	/**
	 * 肪Ȃ̐擪ϊ邩
	 */
	const bool useKanaToOkuri()		{ return okuriHeadTable_ != NULL; }
	
	/**
	 * 肪Ȃ̐擪̕ϊ
	 */
	const UniChar kanaToOkuri(UniChar c) 	{ 
		hash_map<UniChar,UniChar>::iterator i = okuriHeadTable_->find(c);
		if (i == okuriHeadTable_->end())
			return c;
		else
			return i->second;
	}

	/**
	 * ftHg̃CX^XԂ
	 */
	static SkkConfig& config() {
		if (config_)
			return *config_;
		config_ = new SkkConfig();
		return *config_;
	}
	
	// ėpB[Ũz[fBNg̐΃pXԂB
	static string home();
};
