/*  -*- c++ -*-
    $Id: Skkserv.cpp,v 1.2 2003/07/11 10:37:25 phonohawk Exp $
    ---------
	
    MacOS X implementation of the SKK input method.
    Copyright (C) 2002 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "Skkserv.h"
#include "SKKServer.h"
#include "CppCFString.h"
#include <pthread.h>
#include <unistd.h>
#include <iostream>
using namespace std;

static Skkserv* _shared = NULL;
static void listening_thread_begin(Skkserv* serv); // 񃁃o
static void client_thread_begin(Socket* sock); // 񃁃o

Skkserv& Skkserv::sharedServer() {
  if (_shared == NULL) {
    _shared = new Skkserv();
  }
  return *_shared;
}

Skkserv::Skkserv()
  :listening(NULL),local_only(false),flag_term(false) {

}

Skkserv::~Skkserv() {
  stop();
}

Skkserv& Skkserv::setLocalOnly(bool b) {
  local_only = b;
  return *this;
}

Skkserv& Skkserv::start(int port) {
  if (listening != NULL) {
    stop();
  }

  try {
    listening = new Socket("",port,true,20);

    pthread_t pth;
    pthread_create(&pth,NULL,(void*(*)(void*))listening_thread_begin,this);
    pthread_detach(pth);
  }
  catch (string& exception) {
    cerr << exception << endl;

    listening = NULL;
  }

  return *this;
}

Skkserv& Skkserv::stop() {
  if (listening != NULL) {
    flag_term = true;
    while (flag_term) {
      usleep(300);
    }
    
    delete listening;
    listening = NULL;
  }

  return *this;
}

bool Skkserv::isRunning() const {
    return listening != NULL;
}

int Skkserv::port() const {
    if (listening == NULL) {
	return -1;
    }
    else {
	return listening->getPort();
    }
}

static void listening_thread_begin(Skkserv* obj) {
  obj->listening_thread_routine();
}
void Skkserv::listening_thread_routine() {
  while (!flag_term) {
    bool ready_to_accept;
    listening->poll(&ready_to_accept, NULL, NULL, 500);
    if (ready_to_accept) {
      Socket* client = listening->accept();

      if (local_only && client->getpeername() != "127.0.0.1") {
	// 㔋
	delete client;
      }
      else {
	pthread_t pth;
	pthread_create(&pth,NULL,(void*(*)(void*))client_thread_begin,client);
	pthread_detach(pth);
      }
    }
  }
  flag_term = false;
}

static void client_thread_begin(Socket* sock) {
  SkkservSession session(sock);
  session.run();
}

SkkservSession::SkkservSession(Socket* sock)
  : sock(sock) {}

SkkservSession::~SkkservSession() {
  delete sock;
}

void SkkservSession::run() {
    while (1) {
	int command = sock->read();
	if (command == -1) {
	    // ؝
	    command = '0';
	}
    
	if (command == '0') {
	    // ؝
	    break;
	}
	else if (command == '1') {
	    // 
	    string word = sock->readUntil(' ');
	    CppCFString query(word.c_str(),kCFStringEncodingEUC_JP);
	    CppCFString reply(SKKServer::sharedServer().skkserv_style_search(query.trim()));

	    if (reply.length() == 0) {
		// tȂB
		*sock << "4" << word << "\n";
		sock->flush();
	    }
	    else {
		// tB
		*sock << "1" << reply.toCString(kCFStringEncodingEUC_JP) << "\n";
		sock->flush();
	    }
	}
	else if (command == '2') {
	    // T[o[o[W
	    *sock << "AquaSKKServer (skkserv emulation)\n";
	    sock->flush();
	}
	else if (command == '3') {
	    // zXgBt͖ʓ|ȗB
	    *sock << sock->getsockname() << ":" << sock->getsockname() << ":\n";
	    sock->flush();
	}
	else {
	    // Ő㔂؂Ɩʂȉs𑗂ĊzskkinputƂ̒ʐMo҂ȂȂĂ܂ӁB
	    // break;
	}
    }
}
