/*
	$Id: ZenAscInputMode.cpp,v 1.8 2002/12/19 13:37:38 phonohawk Exp $
	---------
	
    MacOS X implementation of the SKK input method.
    Copyright (C) 2002 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "ZenAscInputMode.h"
#include "ParentInputMode.h"
#include "SkkConfig.h"

ZenAscInputMode::ZenAscInputMode(class ParentInputMode& parent)
	:ChildInputMode(parent)
{
	
}

bool ZenAscInputMode::handleInput(unsigned char inputChar)
{
	if (SkkConfig::config().isKakuteiKey(inputChar))
	{
		// ̓[hցB
		parent->goHiraganaInputMode();
		return false;
	}
	else if (SkkConfig::config().isCancelKey(inputChar))
	{
		// do nothing
		return false;
	}
	else
	{
		UniChar c = 0;
		if (inputChar == 0x20)
		{
			c = 0x3000;
		}
		else if (inputChar >= 0x21 && inputChar <= 0x7d)
		{
			c = inputChar - 0x21 + 0xff01;
		}
		else if (inputChar == '~')
		{
			c = 0x301c;
		}
		
		if (c != 0) buffer += c;
		return true;
	}
}

bool ZenAscInputMode::handleBackSpace()
{
	return false;
}

bool ZenAscInputMode::handleReturn()
{
	return false;
}

bool ZenAscInputMode::handleArrow(int kind)
{
	return false;
}

CppCFString ZenAscInputMode::getStringToDisplay()
{
	return buffer;
}

CppCFString ZenAscInputMode::getStringToFix()
{
	if (buffer.length() == 0)
	{
		return buffer;
	}
	else
	{
		CppCFString result = buffer;
		buffer.clear();
		return result;
	}
}

CppCFString ZenAscInputMode::convert(const CppCFString& from)
{
	CppCFString result;
	
	unsigned len = from.length();
	for (unsigned i = 0;i < len;i++)
	{
		UniChar c = from[i];
		if (c == 0x20)
		{
			c = 0x3000;
		}
		else if (c >= 0x21 && c <= 0x7d)
		{
			c = c - 0x21 + 0xff01;
		}
		else if (c == '~')
		{
			c = 0x301c;
		}
		result += c;
	}
	return result;
}

