/*
	$Id: ZenKataInputMode.cpp,v 1.18 2002/12/19 13:37:38 phonohawk Exp $
	---------
	
    MacOS X implementation of the SKK input method.
    Copyright (C) 2002 phonohawk

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include "ZenKataInputMode.h"
#include "ParentInputMode.h"
#include "KanaTreeController.h"
#include "SkkConfig.h"

CharacterTranslator ZenKataInputMode::translator("translation-to-katakana");

ZenKataInputMode::ZenKataInputMode(ParentInputMode& parent)
	:ChildInputMode(parent)
{
	
}

bool ZenKataInputMode::handleInput(const unsigned char inputChar)
{
	// 4siȉ
	if (str_to_fix.length() == 1 && str_to_fix[0] == kIgnoreThis)
	{
		str_to_fix.clear();
	}
	
	if (SkkConfig::config().isToggleKanaKey(inputChar))
	{
		terminate();
		parent->goHiraganaInputMode();
	}
	else if (SkkConfig::config().isToggleKatakanaKey(inputChar))
	{
		terminate();
		parent->goHanKataInputMode();
	}
	else if (SkkConfig::config().isKakuteiKey(inputChar))
	{
		terminate();
	}
	else if (SkkConfig::config().isJisx0208LatinModeKey(inputChar))
	{
		terminate();
		parent->goZenAscInputMode();
	}
	else if (SkkConfig::config().isAbbrevModeKey(inputChar)
                || SkkConfig::config().isLatinModeKey(inputChar))
	{
		// traceĂ݂āAL[/cĂAsciiConvertionModeֈڍsB
		buffer += inputChar;
		str_to_fix += KanaTreeController::defaultTreeController().trace(KanaTreeController::OUTPUT_ZENKATA,buffer);
		
		if (SkkConfig::config().isLatinModeKey(inputChar)
			&& str_to_fix.indexOf(inputChar) != -1)
		{
			terminate();
			str_to_fix.clear();
			parent->goHanAscInputMode();
		}
		else if (SkkConfig::config().isAbbrevModeKey(inputChar)
			&& str_to_fix.indexOf(inputChar) != -1)
		{
			terminate();
			// AsciiConvertionModeֈڍsB
			parent->setEnabledAsciiConvertionMode(true);
			
			// handleInput()IAgetStringToDisplay()kIgnoreThisԂׂɂ
			// str_to_fix()kIgnoreThisĂΗǂBȂƋȁc
			str_to_fix.clear().append(kIgnoreThis);
			return false; // trueԂfixkIgnoreThisB
		}
		
		return str_to_fix.length() > 0;
	}
	else
	{
		buffer += inputChar;
		str_to_fix += KanaTreeController::defaultTreeController().trace(KanaTreeController::OUTPUT_ZENKATA,buffer);
	}
	
	return str_to_fix.length() > 0;
}

bool ZenKataInputMode::handleReturn()
{
	terminate();
	return false;
}

bool ZenKataInputMode::handleBackSpace()
{
	// obt@łȂ΍Ō̕B
	if (buffer.length() > 0)
	{
		buffer.eraseLast(1);
		return true;
	}
	else
	{
		return false;
	}
}

bool ZenKataInputMode::handleArrow(int kind)
{
	terminate();
	return false;
}

bool ZenKataInputMode::handleCg()
{
	if (buffer.length() > 0)
	{
		str_to_fix.clear();
		buffer.clear();
		return true;
	}
	else
	{
		return false;
	}
}

bool ZenKataInputMode::isQueueEmpty() const
{
	return buffer.length() == 0;
}

void ZenKataInputMode::terminate()
{
	// obt@łȂAŌ̕n
	// str_to_fixɁuvǉB
	// ǂɂĂobt@͋ɂB
	if (SkkConfig::config().fixN() && buffer.length() > 0 && buffer[0] == 'n')
	{
		str_to_fix.append(0x30f3); // 
	}
	buffer.clear();
}

CppCFString ZenKataInputMode::getStringToDisplay()
{
	if (str_to_fix.length() > 0)
	{
		return str_to_fix + buffer;
	}
	else
	{
		return buffer;
	}
}

CppCFString ZenKataInputMode::getStringToFix()
{
	if (str_to_fix.length() == 0)
	{
		return str_to_fix;
	}
	else
	{
		CppCFString str_to_return(str_to_fix);
		str_to_fix.clear();
		
		return str_to_return;
	}
}

CppCFString ZenKataInputMode::convert(const CppCFString& from)
{
	return translator.translate(from);
}
